import ../types/value
import ../types/native
import bitops
import math
import strformat

# vararg math funcs:

# 2 arg math funcs:

proc natPow(args: seq[NdValue], retval: var NdValue): NatReturn =
  if args.len() != 2:
    return natError(&"Power expects 2 arguments, got {args.len()}.")
  let base = args[0]
  let exp = args[1]
  if base.isFloat() and exp.isFloat():
    retval = base.asFloat().pow(exp.asFloat()).fromFloat();
    return natOk
  return natError(&"Power expects 2 numbers, however got {base.friendlyType()} and {exp.friendlyType()}.")

proc natMod(args: seq[NdValue], retval: var NdValue): NatReturn =
  if args.len() != 2:
    return natError(&"Mod expects 2 arguments, got {args.len()}.")
  let base = args[0]
  let exp = args[1]
  if base.isFloat() and exp.isFloat():
    let modulo = base.asFloat().int() mod exp.asFloat().int()
    retval = modulo.float().fromFloat()
    return natOk
  return natError(&"Mod expects 2 numbers, however got {base.friendlyType()} and {exp.friendlyType()}.")


proc constructMath* =
  # vararg funcs:
  #defNative("min", natMin)
  #defNative("max", natMax)

  # 2 arg funcs:
  defNative("pow", natPow)
  defNative("mod", natMod)

  # 1 arg funcs:
  #defNative("ln", natLn)
  #defNative("floor", natFloor)
  #defNative("sin", natSin)
  #defNative("cos", natCos)
  #defNative("tan", natTan)

  # 0 arg funcs:
  #defNative("e", natE)
  #defNative("pi", natPI)

